/*
 * IRQ-handling code for TiVo devices
 *
 * Copyright (C) 1992 Linus Torvalds
 * Copyright (C) 1994 - 2000 Ralf Baechle
 * Copyright (C) 2000-2003 TiVo Inc. All Rights Reserved.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel_stat.h>
#include <linux/slab.h>
#include <linux/random.h>
#include <linux/plog.h>
#include <linux/module.h>
#include <linux/sw_watchdog.h>

#include <asm/mipsregs.h>
#include <asm/ptrace.h>
#include <asm/bcm7315.h>
#include <asm/atomic.h>

extern asmlinkage void tivo_handle_irq(void);
 
static volatile u8* const bcm_reg_base = __PERIPHREG(0);

#define GIO_RO(r)   (BCM_GIO_##r - __PERIPHREG(0))
#define GIOX_RO(r)  (BCM_GIOEXT_##r - __PERIPHREG(0))
#define MGIO_RO(r)  (BCM_MGIO_##r - __PERIPHREG(0))

#define GIO_BYTES	9

static const u8 gio_reg_offsets[GIO_BYTES][3] = {
    { GIO_RO(MASKLSB),   GIO_RO(RSTLSB),   GIO_RO(STATLSB)   },
    { GIO_RO(MASKHSB),   GIO_RO(RSTHSB),   GIO_RO(STATHSB)   },
    { GIO_RO(MASK2MSB),  GIO_RO(RST2MSB),  GIO_RO(STAT2MSB)  },
    { GIO_RO(MASKMSB),   GIO_RO(RSTMSB),   GIO_RO(STATMSB)   },
    { MGIO_RO(MASKLSB),  MGIO_RO(RSTLSB),  MGIO_RO(STATLSB)  },
    { MGIO_RO(MASKHSB),  MGIO_RO(RSTHSB),  MGIO_RO(STATHSB)  },
    { MGIO_RO(MASK2MSB), MGIO_RO(RST2MSB), MGIO_RO(STAT2MSB) },
    { MGIO_RO(MASKMSB),  MGIO_RO(RSTMSB),  MGIO_RO(STATMSB)  },
    { GIOX_RO(MASKLSB),  GIOX_RO(RSTLSB),  GIOX_RO(STATLSB)  },
};

#undef GIO_RO
#undef GIOX_RO
#undef MGIO_RO

#define GIO_MASK	0
#define GIO_RST		1
#define GIO_STAT	2

#define GIO_REG(byte, ireg)	(bcm_reg_base + gio_reg_offsets[byte][ireg])

atomic_t irq_err_count;

static struct irqaction *irq_action[BCM_NR_IRQS];
static int irq_depth[BCM_NR_IRQS];

#ifdef CONFIG_TIVO_DEVEL

int debug_irqs;

static int __init dbgirq_setup(char *str)
{
	get_option(&str, &debug_irqs);
	return 1;
}

__setup("dbgirq=", dbgirq_setup);

#define DEBUG_IRQS debug_irqs

#else
#define DEBUG_IRQS 0
#endif

static inline void mask_irq(unsigned int irq)
{
	unsigned int status;

	if (irq < 1) {
		// Nothing for IRQ 0
	} else if (irq < 33) {
		*BCM_INTMASK &= ~(1 << (irq - 1));
	} else if (irq < 37) {
		// Nothing for IRQs 33-36
	} else if (irq < 45) {
		*BCM_PM1_IRQEN_LO &= ~(1 << (irq - 37));
	} else if (irq < 53) {
		*BCM_PM1_IRQEN_HI &= ~(1 << (irq - 45));
	} else if (irq < 62) {
		status = read_32bit_cp0_register(CP0_STATUS);
		status &= ~(1 << (irq - 53 + 8));
		write_32bit_cp0_register(CP0_STATUS, status);
#ifdef CONFIG_SW_WATCHDOG
		_gbl_sti_cli_flags &= ~(1 << (irq - 53 + 8));
#endif
	} else {
		int byte = (irq - BCM_LINUX_GIO_IRQ_BASE) / 8;
		int bit = (irq - BCM_LINUX_GIO_IRQ_BASE) & 7;
		*GIO_REG(byte, GIO_MASK) &= ~(1 << bit);
	}
	barrier();
}

static inline void mask_and_ack_irq(unsigned int irq)
{
	if (irq < 62) {
		return mask_irq(irq);
	} else {
		int byte = (irq - BCM_LINUX_GIO_IRQ_BASE) / 8;
		int bit = (irq - BCM_LINUX_GIO_IRQ_BASE) & 7;
		*GIO_REG(byte, GIO_MASK) &= ~(1 << bit);
		*GIO_REG(byte, GIO_RST) = 1 << bit;
		barrier();
		*GIO_REG(byte, GIO_RST) = 0;
		barrier();
	}
}

static inline void ack_irq(unsigned int irq)
{
	if (irq >= 62) {
		int byte = (irq - BCM_LINUX_GIO_IRQ_BASE) / 8;
		int bit = (irq - BCM_LINUX_GIO_IRQ_BASE) & 7;
		*GIO_REG(byte, GIO_RST) = 1 << bit;
		barrier();
		*GIO_REG(byte, GIO_RST) = 0;
	}
	barrier();
}

static inline void unmask_irq(unsigned int irq)
{
	unsigned int status;

	if (irq < 1) {
		// Nothing for IRQ 0
	} else if (irq < 33) {
		*BCM_INTMASK |= 1 << (irq - 1);
	} else if (irq < 37) {
		// Nothing for IRQs 33-36
	} else if (irq < 45) {
		*BCM_PM1_IRQEN_LO |= 1 << (irq - 37);
	} else if (irq < 53) {
		*BCM_PM1_IRQEN_HI |= 1 << (irq - 45);
	} else if (irq < 62) {
		status = read_32bit_cp0_register(CP0_STATUS);
		status |= 1 << (irq - 53 + 8);
		write_32bit_cp0_register(CP0_STATUS, status);
#ifdef CONFIG_SW_WATCHDOG
		_gbl_sti_cli_flags |= 1 << (irq - 53 + 8);
#endif
	} else {
		int byte = (irq - BCM_LINUX_GIO_IRQ_BASE) / 8;
		int bit = (irq - BCM_LINUX_GIO_IRQ_BASE) & 7;
		*GIO_REG(byte, GIO_MASK) |= 1 << bit;
	}
	barrier();
}

void disable_irq(unsigned int irq_nr)
{
	unsigned long flags;

	if (DEBUG_IRQS >= 3) {
		printk("dis<%d>", irq_nr);
	}
	if (irq_nr >= BCM_NR_IRQS)
		return;
	save_and_cli(flags);
	if (irq_depth[irq_nr]++ < 0)
		mask_irq(irq_nr);
	restore_flags(flags);
}

void enable_irq(unsigned int irq_nr)
{
	unsigned long flags;

	if (DEBUG_IRQS >= 3) {
		printk("ena<%d>", irq_nr);
	}
	if (irq_nr >= BCM_NR_IRQS)
		return;
	save_and_cli(flags);
	if (--irq_depth[irq_nr] < 0)
		unmask_irq(irq_nr);
	restore_flags(flags);
}

EXPORT_SYMBOL(disable_irq);
EXPORT_SYMBOL(enable_irq);

int get_irq_list(char *buf)
{
	int i, len = 0;
	struct irqaction *action;

	for (i = 0 ; i < BCM_NR_IRQS ; i++) {
		action = irq_action[i];
		if (!action)
			continue;
		len += sprintf(buf + len, "%2d: %8d %c %s",
			i, kstat.irqs[0][i],
			(action->flags & SA_INTERRUPT) ? '+' : ' ',
			action->name);
		for (action = action->next; action; action = action->next) {
			len += sprintf(buf + len, ",%s %s",
				(action->flags & SA_INTERRUPT) ? " +" : "",
				action->name);
		}
		len += sprintf(buf + len, "\n");
	}
	return len;
}

/*
 * do_IRQ is called by the low-level interrupt handling code with the
 * appropriate IRQ number
 */
static void do_IRQ(int irq, struct pt_regs *regs)
{
	struct irqaction *action;
	int cpu;
	mask_and_ack_irq(irq);
	cpu = smp_processor_id();
	irq_enter(cpu, irq);
#ifdef CONFIG_SW_WATCHDOG
	/* Records first level IRQs for logging */
	sw_watchdog_irq_enter(cpu, irq); 
#endif
	kstat.irqs[cpu][irq]++;

	action = irq_action[irq];

	if (DEBUG_IRQS >= 2) {
		if (irq == BCM_LINUX_TIMR_IRQ && DEBUG_IRQS < 5) {
			static unsigned int count1;
			if (!(count1++ & 0xff)) {
				printk("*");
			}
		} else if (irq == BCM_LINUX_BGE_IRQ && DEBUG_IRQS < 5) {
			static unsigned int count2;
			if (!(count2++ & 0xff)) {
				printk("x");
			}
		} else if (irq == BCM_LINUX_USB_HOST_IRQ && DEBUG_IRQS < 5) {
			static unsigned int count3;
			if (!(count3++ & 0xff)) {
				printk("+");
			}
		} else {
			printk("[%d", irq);
		}
	}

	if (action != 0) {
		unsigned long flags = 0;

		plog_intr_enter(irq, regs->cp0_epc);
		do {
#ifdef CONFIG_SW_WATCHDOG
			/* WD Masks out everything but wrap count: */
			__cli(); 
			/* WD Turns back on interupt enable bit: */
			__really_sti();
#endif

			if (!(action->flags & SA_INTERRUPT))
				__sti();
			action->handler(irq, action->dev_id, regs);
			__cli();
			flags |= action->flags;
			action = action->next;
		} while (action);
		if (flags & SA_SAMPLE_RANDOM)
			add_interrupt_randomness(irq);
		// if IRQ was disabled from within its own handler(s),
		// leave it masked, otherwise re-enable
		if (irq_depth[irq] < 0)
			unmask_irq(irq);
		plog_intr_leave(irq);
	} else if (DEBUG_IRQS >= 2) {	// report as unhandled
		printk("!");
	}
	if (DEBUG_IRQS >= 2) {
		if ((irq != BCM_LINUX_TIMR_IRQ &&
		     irq != BCM_LINUX_BGE_IRQ &&
		     irq != BCM_LINUX_USB_HOST_IRQ) ||
                    DEBUG_IRQS >= 5) {
			printk("]");
		}
	}

	irq_exit(cpu, irq);

	if (softirq_pending(cpu))
		do_softirq();
#ifdef CONFIG_SW_WATCHDOG
	/* Make CP0_STATUS sane again... */
	{
		unsigned int cp0status = read_32bit_cp0_register(CP0_STATUS);
		cp0status &= ~(1); /* Clear interrupt enabled */
		cp0status |= (_gbl_sti_cli_flags & 0x7f00);
		write_32bit_cp0_register(CP0_STATUS, cp0status);
	}
#endif
}

int tivo_setup_irq(int irq, struct irqaction *new)
{
	int shared = 0;
	struct irqaction *old, **p;
	unsigned long flags;

	p = irq_action + irq;
	if ((old = *p) != NULL) {
		// Shared ints must all agree to share
		if (!(old->flags & new->flags & SA_SHIRQ))
			return -EBUSY;
		do {
			p = &old->next;
			old = *p;
		} while (old);
		shared = 1;
	}
	if (new->flags & SA_SAMPLE_RANDOM)
		rand_initialize_irq(irq);

	save_and_cli(flags);
	*p = new;

	if (!shared) {
		irq_depth[irq] = -1;
		ack_irq(irq);
		unmask_irq(irq);
	}
	restore_flags(flags);
	return 0;
}

int request_irq(unsigned int irq,
	void (*handler)(int, void *, struct pt_regs *),
	unsigned long irqflags,
	const char *devname,
	void *dev_id)
{
	int retval;
	struct irqaction *action;

	if (DEBUG_IRQS >= 1) {
		printk("request irq %d\n", irq);
	}
	if (irq >= BCM_NR_IRQS)
		return -EINVAL;
	if (!handler)
		return -EINVAL;

	action = (struct irqaction *)kmalloc(sizeof(struct irqaction), GFP_KERNEL);
	if (!action)
		return -ENOMEM;

	action->handler = handler;
	action->flags = irqflags;
	action->mask = 0;
	action->name = devname;
	action->next = NULL;
	action->dev_id = dev_id;

	retval = tivo_setup_irq(irq, action);

	if (retval)
		kfree(action);
	return retval;
}

void free_irq(unsigned int irq, void *dev_id)
{
	struct irqaction *action, **p;
	unsigned long flags;

	if (DEBUG_IRQS >= 1) {
		printk("free irq %d\n", irq);
	}
	if (irq >= BCM_NR_IRQS) {
		printk(KERN_ERR "Error - trying to free invalid IRQ %d\n", irq);
		return;
	}
	for (p = irq + irq_action; (action = *p) != NULL; p = &action->next) {
		if (action->dev_id != dev_id)
			continue;

		// Found it - now free it
		save_and_cli(flags);
		*p = action->next;
		if (!irq[irq_action])
			mask_irq(irq);
		restore_flags(flags);
		kfree(action);
		return;
	}
	printk(KERN_ERR "Trying to free free IRQ%d\n", irq);
}

unsigned long probe_irq_on(void)
{
	// not implemented
	return 0;
}

unsigned int probe_irq_mask(unsigned long val)
{
	// not implemented
	return 0;
}

int probe_irq_off(unsigned long irqs)
{
	// not implemented
	return 0;
}

int (*irq_cannonicalize)(int irq);

static int bcm_irq_cannonicalize(int irq)
{
	return irq;
}

static struct irqaction cascade =
	 { NULL, SA_INTERRUPT, 0, "cascade", NULL, NULL };

void __init init_IRQ(void)
{
	int i;
	unsigned int status;

	irq_cannonicalize = bcm_irq_cannonicalize;
	set_except_vector(0, tivo_handle_irq);

	status = read_32bit_cp0_register(CP0_STATUS);
	status &= 0xfffe00ff;
	write_32bit_cp0_register(CP0_STATUS, status);

	// Disable all BCM7315 ints
	*BCM_INTMASK = 0;
	*BCM_PM1_IRQEN_LO = 0;
	*BCM_PM1_IRQEN_HI = 0;
	for (i = 0; i < GIO_BYTES; i++) {
		*GIO_REG(i, GIO_MASK) = 0;
	}
	barrier();

	tivo_setup_irq(BCM_LINUX_INTERNAL_IRQ, &cascade);
	tivo_setup_irq(BCM_LINUX_UPG_IRQ, &cascade);
	tivo_setup_irq(BCM_LINUX_UPG_GIO_IRQ, &cascade);
}

static inline void do_gio_irq(struct pt_regs *regs)
{
	int j,i;
	for (j = 0; j < GIO_BYTES; j++) {
		u8 gio_stat = *GIO_REG(j, GIO_STAT) & *GIO_REG(j, GIO_MASK);

		if (gio_stat) {
			for (i = 0; i < 8; i++) {
				if (gio_stat & (1 << i)) {
					do_IRQ(BCM_LINUX_GIO_IRQ_BASE +
					       j * 8 + i, regs);
					return;
				}
			}
		}
	}
	if (DEBUG_IRQS >= 1) {
		printk(KERN_WARNING "spurious GIO int!\n");
	}
	atomic_inc(&irq_err_count);
}

static inline void do_upg_irq(struct pt_regs *regs)
{
	int i;	
	u8 upg_stat;
	
	upg_stat = *BCM_PM1_IRQSTAT_HI & *BCM_PM1_IRQEN_HI;
	if (upg_stat) {
		for (i = 0; i <= 3; i++) {
			if (upg_stat & (1 << i)) {
				if (i == BCM_LINUX_UPG_GIO_IRQ - 45) {
					do_gio_irq(regs);
					return;
				}
				do_IRQ(i + 45, regs);
				return;
			}
		}
	} else {
		upg_stat = *BCM_PM1_IRQSTAT_LO & *BCM_PM1_IRQEN_LO;
		for (i = 0; i <= 7; i++) {
			if (upg_stat & (1 << i)) {
				do_IRQ(i + 37, regs);
				return;
			}
		}
	}
	if (DEBUG_IRQS >= 1) {
		printk(KERN_WARNING "spurious UPG int!\n");
	}
	atomic_inc(&irq_err_count);
}

static inline void do_7315_irq(struct pt_regs *regs)
{
	int i;
	u32 intr_stat = *BCM_INTSTATUS & *BCM_INTMASK;

	for (i = 31; i >= 0; i--) {
		if (intr_stat & (1 << i)) {
			if (i == BCM_LINUX_UPG_IRQ - 1) {
				do_upg_irq(regs);
				return;
			}
			do_IRQ(i + 1, regs);
			return;
		}
	}
	if (DEBUG_IRQS >= 1) {
		printk(KERN_WARNING "spurious BCM int! BCM_INTSTATUS=0x%08x\n", *BCM_INTSTATUS);
	}
	atomic_inc(&irq_err_count);
}

asmlinkage void do_vector(u32 irqmask, struct pt_regs *regs)
{
	int i;

	for (i = 0; i < 9; i++) {
		if (irqmask & (1 << (i + 8))) {
			if (i == 2) {
				do_7315_irq(regs);
				return;
			}
			do_IRQ(i + 53, regs);
			return;
		}
	}

	if (DEBUG_IRQS >= 4) {
	// This was slowing down the local bus UARTs' interrupt too much,
	// causing serial port FIFO overflow (and thus crypto failure).
	// It needs to be investigated why we're getting spurious ints,
	// but only spew this warning if DEBUG_IRQ for now...
		printk(KERN_WARNING "do_vector got spurious int with mask 0x%08x!\n", irqmask);
	}
	atomic_inc(&irq_err_count);
}

void init_irq_proc (void)
{
}
